from typing import Dict, Union

import cv2
import easyocr
import numpy as np


class TextExtraction:
    def __init__(self, languages=['vi', 'en'], use_gpu=False):
        self.reader = easyocr.Reader(languages, gpu=use_gpu)

    def extract_text(self, image_source: Union[str, np.ndarray]) -> Dict:
        assert isinstance(image_source, np.ndarray) or isinstance(
            image_source, str
        ), "Image source must be a string or a numpy array."

        if isinstance(image_source, str):
            image = cv2.imread(image_source)

        else:
            image = image_source

        results = self.reader.readtext(image)
        final_results = []
        for bbox, text, _ in results:
            final_results.append(
                {
                    'bbox': [
                        int(bbox[0][0]),
                        int(bbox[0][1]),
                        int(bbox[2][0]),
                        int(bbox[2][1]),
                    ],
                    'text': text,
                    'center': [
                        int((bbox[0][0] + bbox[2][0]) / 2),
                        int((bbox[0][1] + bbox[2][1]) / 2),
                    ],
                }
            )

        return final_results



from typing import Dict, List

from fitz import Rect

from ..const import BorderFormat, HorizontalAlign, OutputFormat, VerticalAlign


class TableReconstruction:
    def __init__(self) -> None:
        pass

    def process(
        self,
        table_structure: Dict,
        table_values: Dict,
        output_formats: List[OutputFormat] = [
            OutputFormat.HTML,
            OutputFormat.LATEX,
            OutputFormat.CSV,
        ],
        border_format: BorderFormat = BorderFormat.FULL_BORDER,
        vertical_align: VerticalAlign = VerticalAlign.MIDDLE,
        horizontal_align: HorizontalAlign = HorizontalAlign.LEFT,
    ) -> Dict:
        """Process table components and values to reconstruct table

        Args:
            table_structure (Dict): Table components detected by table structure recognition module
            table_values (Dict): Table values detected by text extraction module
            output_format (List[OutputFormat]): List of output formats to reconstruct table
            border_format (BorderFormat, optional): Border format of the table. Defaults to BorderFormat.FULL_LINE.
            vertical_align (VerticalAlign, optional): Vertical align of the table. Defaults to VerticalAlign.MIDDLE.
            horizontal_align (HorizontalAlign, optional): Horizontal align of the table. Defaults to HorizontalAlign.LEFT.

        Returns:
            Dict: Dictionary of outputs and table cells.
        """
        ## Verify output formats, border format, vertical align, horizontal align
        all_output_formats = [i for i in OutputFormat]
        all_border_formats = [i for i in BorderFormat]
        all_vertical_aligns = [i for i in VerticalAlign]
        all_horizontal_aligns = [i for i in HorizontalAlign]

        assert all(
            output_format in all_output_formats for output_format in output_formats
        ), f"Invalid output format: {output_formats}"
        assert (
            border_format in all_border_formats
        ), f"Invalid border format: {border_format}"
        assert (
            vertical_align in all_vertical_aligns
        ), f"Invalid vertical align: {vertical_align}"
        assert (
            horizontal_align in all_horizontal_aligns
        ), f"Invalid horizontal align: {horizontal_align}"

        ## Get cells from table structure
        table_cells = table_structure["cells"]
        num_rows = table_structure["num_rows"]
        num_columns = table_structure["num_columns"]
        table_cells = self._fill_values(table_cells, table_values)
        table_cells = self._group_values(table_cells)

        ## Format output
        output = {}
        for output_format in output_formats:
            output[output_format] = self._format_output(
                table_cells,
                num_rows,
                num_columns,
                output_format,
                border_format,
                vertical_align,
                horizontal_align,
            )

        return {
            "outputs": output,
            "table_cells": table_cells,
            "num_rows": num_rows,
            "num_cols": num_columns,
        }

    def _fill_values(
        self, table_cells: List[Dict], table_values: List[Dict]
    ) -> List[Dict]:
        """Fill values from TE module into table structure from TSR module

        Args:
            table_cells (List[Dict]): List of cells in table. Cells are represented by dictionary with keys: bbox, column_nums, row_nums
            table_values (List[Dict]): List of values detected by text extraction module. Values are represented by dictionary with keys: bbox, text, center

        Returns:
            List[Dict]: List of cells in table with values filled. Cells are represented by dictionary with keys: bbox, column_nums, row_nums, values
        """
        sorted_instances = sorted(
            table_values, key=lambda x: (x['center'][0], x['center'][1])
        )
        for cell in table_cells:
            cell['values'], remaining_values = [], []

            for value in sorted_instances:
                if Rect(cell['bbox']).contains(value['center']):
                    cell['values'].append(value)
                else:
                    remaining_values.append(value)
            sorted_instances = remaining_values.copy()
        return table_cells

    def _group_values(self, table_cells: List[Dict]) -> List[Dict]:
        """Group values in the same cell into a single cell content

        Args:
            table (List[Dict]): table structure with values filled (not merged yet). Format:
                [
                    {
                        'bbox': [x1, y1, x2, y2],
                        'column_nums': [column_num1, column_num2, ...],
                        'row_nums': [row_num1, row_num2, ...],
                        'values': [
                            {
                                'bbox': [x1, y1, x2, y2],
                                'text': 'value',
                                'center': [x, y],
                            },
                            ...
                        ],
                    },
                    ...
                ]
            merge_dist (int, optional): Maximum distance between two values to be merged. Defaults to 10.

        Returns:
            List[Dict]: table structure with values grouped. Format:
                [
                    {
                        'bbox': [x1, y1, x2, y2],
                        'column_nums': [column_num1, column_num2, ...],
                        'row_nums': [row_num1, row_num2, ...],
                        'content': 'cell content',
                    },
                    ...
                ]
        """

        for cell in table_cells:
            if len(cell["values"]) == 0 or len(cell["values"]) == 1:
                cell["content"] = (
                    cell["values"][0]['text'] if len(cell["values"]) == 1 else ""
                )
                cell.pop("values", None)
                continue
            sorted_values = sorted(
                cell["values"], key=lambda x: (x['center'][1], x['center'][0])
            )

            line = 1
            pre_value = sorted_values[0]
            pre_value["line"] = line

            grouped_values = [pre_value]

            for value in sorted_values[1:]:
                if (
                    value["bbox"][1] < pre_value["center"][1]
                    and value["bbox"][3] > pre_value["center"][1]
                ):
                    value["line"] = line
                else:
                    line += 1
                    value["line"] = line
                    pre_value = value
                grouped_values.append(value)

            grouped_values = sorted(
                grouped_values, key=lambda x: (x["line"], x["center"][0])
            )

            pre_line = 1
            cell["content"] = grouped_values[0]["text"]
            for value in grouped_values[1:]:
                if value["line"] == pre_line:
                    cell["content"] += f" {value['text']}"
                else:
                    cell["content"] += f"\n{value['text']}"
                pre_line = value['line']

            cell.pop("values", None)
        return table_cells

    def _format_output(
        self,
        table_cells: List[Dict],
        num_rows: int,
        num_columns: int,
        output_format: OutputFormat,
        border_format: BorderFormat,
        vertical_align: VerticalAlign,
        horizontal_align: HorizontalAlign,
    ) -> str:
        """Format table to output format

        Args:
            table_cells (List[Dict]): table structure with values grouped after filling and grouping values of eachs cells.
            num_rows (int): Number of rows in table
            num_columns (int): Number of columns in table
            output_format (OutputFormat): Output format to convert table to
            border_format (BorderFormat): Border format of the table
            vertical_align (VerticalAlign): Vertical align of the table
            horizontal_align (HorizontalAlign): Horizontal align of the table

        Returns:
            str: Table content in output format
        """
        if output_format == OutputFormat.CSV:
            return self._to_csv(table_cells, num_rows, num_columns)
        elif output_format == OutputFormat.HTML:
            return self._to_html(
                table_cells,
                num_rows,
                num_columns,
                border_format,
                vertical_align,
                horizontal_align,
            )
        elif output_format == OutputFormat.LATEX:
            return self._to_latex(
                table_cells,
                num_rows,
                num_columns,
                border_format,
                vertical_align,
                horizontal_align,
            )

    def _to_csv(self, table_cells: List[Dict], num_rows: int, num_columns: int) -> str:
        """Convert table to csv format

        Args:
            table_cells (List[Dict]): table structure with values grouped after filling and grouping values of eachs cells.
            num_rows (int): number of rows in table
            num_columns (int): number of columns in table

        Returns:
            str: CSV string of table
        """
        ## Create an empty matrix
        output = [["" for _ in range(num_columns)] for _ in range(num_rows)]

        for cell in table_cells:
            row_num = cell["row_nums"][0]
            column_num = cell["column_nums"][0]
            content = cell["content"]
            if (
                content.find(",") != -1
                or content.find('\n') != -1
                or content.find('"') != -1
            ):
                content = content.replace("\n", "\\n")
                content = content.replace('"', "''")
                content = f'"{content}"'
            output[row_num][column_num] = content

        ## Convert to csv format with each row is a line, each cell is separated by comma
        output = [",".join(row) for row in output]
        output = "\n".join(output)
        return output

    def _to_html(
        self,
        table_cells: List[Dict],
        num_rows: int,
        num_columns: int,
        border_format: BorderFormat,
        vertical_align: VerticalAlign,
        horizontal_align: HorizontalAlign,
    ) -> str:
        """Convert table to html format

        Args:
            table_cells (List[Dict]): table structure with values grouped after filling and grouping values of eachs cells.
            num_rows (int): number of rows in table
            num_columns (int): number of columns in table
            border_format (BorderFormat): Border format of the table
            vertical_align (VerticalAlign): Vertical align of the table
            horizontal_align (HorizontalAlign): Horizontal align of the table

        Returns:
            str: HTML code of table
        """

        ## Step 1: Create HTML table
        i_row, i_col = 0, 0
        script_table = ""

        while i_row < num_rows:
            i_col = 0
            text_row = ""
            while i_col < num_columns:
                text_cell = ""
                for cell in table_cells:
                    # coors here mean index (row, col) in table
                    col_nums = cell['column_nums']
                    row_nums = cell['row_nums']

                    # check if it is the last cell in the row (i_row)
                    if (col_nums[0] == i_col) and (row_nums[0] == i_row):
                        rowspan = len(row_nums)
                        colspan = len(col_nums)

                        content = cell["content"]
                        content = (
                            content.replace("<", "&lt;")
                            .replace(">", "&gt;")
                            .replace("&", "&amp;")
                        )
                        content = content.replace("\n", "<br>")

                        text_cell = "<td rowspan='{}', colspan='{}'>{}</td>".format(
                            str(rowspan), str(colspan), content
                        )
                text_row += text_cell
                i_col += 1

            text_row = "<tr>{}</tr>".format(text_row)
            script_table += text_row
            i_row += 1

        script_table = "<table>{}</table>".format(script_table)

        script_table = (
            script_table.replace("\"http", "'http")
            .replace("MathML\"", "MathML'")
            .replace("\"inline\"", "'inline'")
        )

        ## Step 2: Add style to HTML table
        ### Border format
        border_map = {
            BorderFormat.FULL_BORDER: "table, td {border-collapse: collapse; border: 1px solid black;}",
            BorderFormat.NO_BORDER: "table {border-collapse: collapse;}",
            BorderFormat.HOR_BORDER: "table {border-collapse: collapse;} td {border-top: 1px solid black; border-bottom: 1px solid black;}",
            BorderFormat.VER_BORDER: "table {border-collapse: collapse;} td {border-left: 1px solid black; border-right: 1px solid black;}",
            BorderFormat.INNER_BORDER: "table {border-collapse: collapse; border-style: hidden;} td {border: 1px solid black;}",
            BorderFormat.OUTER_BORDER: "table {border-collapse: collapse; border: 1px solid black;} td {border: none;}",
        }

        border = border_map[border_format]
        ### Align format
        align = (
            f"td {{text-align: {horizontal_align}; vertical-align: {vertical_align};}}"
        )

        style = f"<style> table {{table-layout: fixed; width: 100%;}} {border} {align}</style>"

        html_code = f"{style} {script_table}"

        return html_code

    def _to_latex(
        self,
        table_cells: List[Dict],
        num_rows: int,
        num_columns: int,
        border_format: BorderFormat,
        vertical_align: VerticalAlign,
        horizontal_align: HorizontalAlign,
    ) -> str:
        """Convert table to latex format

        Args:
            table_cells (List[Dict]): table structure with values grouped after filling and grouping values of eachs cells.
            num_rows (int): number of rows in table
            num_columns (int): number of columns in table
            border_format (BorderFormat): Border format of the table
            vertical_align (VerticalAlign): Vertical align of the table
            horizontal_align (HorizontalAlign): Horizontal align of the table

        Returns:
            str: Latex code of table
        """
        v_align_map = {
            VerticalAlign.TOP: "h",
            VerticalAlign.MIDDLE: "m",
            VerticalAlign.BOTTOM: "f",
        }
        h_align_map = {
            HorizontalAlign.LEFT: "l",
            HorizontalAlign.CENTER: "c",
            HorizontalAlign.RIGHT: "r",
        }

        ## Step 1: Create Latex table
        structure = ""
        for cell in table_cells:
            r = cell["row_nums"][0] + 1
            c = cell["column_nums"][0] + 1
            n_c = len(cell['column_nums'])
            n_r = len(cell['row_nums'])
            structure += (
                "cell{"
                + str(r)
                + "}{"
                + str(c)
                + "}= {c="
                + str(n_c)
                + ", r="
                + str(n_r)
                + "}{"
                + h_align_map[horizontal_align]
                + ","
                + v_align_map[vertical_align]
                + "},"
            )

        table_value = ""
        i_row, i_col = 0, 0
        while i_row < num_rows:
            i_col = 0
            text_row = ""
            while i_col < num_columns:
                text_cell = ""
                for cell in table_cells:
                    col_nums = cell['column_nums']
                    row_nums = cell['row_nums']
                    if (col_nums[0] == i_col) and (row_nums[0] == i_row):
                        content = cell["content"]
                        content = (
                            content.replace("\\", "$\\backslash$")
                            .replace("#", "\\#")
                            .replace("$", "\\$")
                            .replace("%", "\\%")
                            .replace("&", "\\&")
                            .replace("_", "\\_")
                            .replace("~", "\\textasciitilde{}")
                            .replace("^", "\\textasciicircum{}")
                            .replace("}", "\\}")
                            .replace("{", "\\{")
                        )
                        content = content.replace("\n", "\\\\")
                        text_cell = "{" + content + "}"
                text_row += text_cell + "&"
                i_col += 1

            text_row = text_row[:-1]
            text_row = f"{text_row}\\\\"
            table_value += text_row
            i_row += 1

        ## Border format
        border_map = {
            BorderFormat.FULL_BORDER: "hlines, vlines, ",
            BorderFormat.NO_BORDER: "",
            BorderFormat.HOR_BORDER: "hlines, ",
            BorderFormat.VER_BORDER: "vlines, ",
            BorderFormat.INNER_BORDER: "hlines, vlines, hline{1,Z} = {0pt}, vline{1,Z} = {0pt}, ",
            BorderFormat.OUTER_BORDER: "hline{1,Z} = {1pt,solid}, vline{1,Z} = {1pt,solid}, ",
        }
        border = border_map[border_format]

        script_table = (
            "\\begin{tblr}{" + border + structure + "} " + table_value + "\n\\end{tblr}"
        )

        doc_script = (
            "\\documentclass{article}\n\\usepackage{geometry}\n\\geometry{a4paper,margin=2mm}\n\\usepackage{tabularray}\n\\usepackage[utf8]{vietnam}\n\\begin{document}\\begin{center}\n"
            + script_table
            + "\n\\end{center}\n\\end{document}"
        )

        return doc_script



from typing import Dict, List, Union

import numpy as np
from PIL import Image

from .const import BorderFormat, HorizontalAlign, OutputFormat, VerticalAlign
from .table_detection.table_detection import TableDetection
from .table_reconstruction.table_reconstruction import TableReconstruction
from .table_structure_recognition.table_structure_recognition import (
    TableStructureRecognition,
)
from .text_extraction.text_extraction import TextExtraction


class TableExtraction:

    def __init__(
        self, td_weight_path: str = None, tsr_weight_path: str = None, device="cpu"
    ):
        """Initialize TableExtraction class.

        Args:
            td_weight_path (str, optional): Path to table detection weight. Defaults to None (use default weight).
            tsr_weight_path (str, optional): Path to table structure recognition weight. Defaults to None (use default weight).
            device (str, optional): Device to run the model. Defaults to "cpu".
        """

        ## Initialize components
        self.table_detection = TableDetection(weight_path=td_weight_path, device=device)
        self.table_structure_recognition = TableStructureRecognition(
            weight_path=tsr_weight_path, device=device
        )

        ## NOTE: EasyOCR only run on card 0, to prevent error, set use_gpu=False
        self.text_extraction = TextExtraction(
            # use_gpu=False if device == "cpu" else True
            use_gpu=False
        )
        self.table_reconstruction = TableReconstruction()
        self.default_hparams = {
            "td": {
                "conf": 0.7,
                "iou": 0.4,
                "extend_ratio": 0.12,
                "padding": 40,
            },
            "tsr": {
                "conf": 0.25,
                "iou": 0.5,
            },
        }

    def extract_table(
        self,
        image_source: Union[str, np.ndarray, Image.Image],
        output_formats: List[OutputFormat] = [
            OutputFormat.CSV,
            OutputFormat.HTML,
            OutputFormat.LATEX,
        ],
        border_format: BorderFormat = BorderFormat.FULL_BORDER,
        vertical_align: VerticalAlign = VerticalAlign.MIDDLE,
        horizontal_align: HorizontalAlign = HorizontalAlign.LEFT,
    ) -> List[Dict]:
        """Extract tables from the input image source.

        Args:
            image_source (Union[str, np.ndarray, Image.Image]): Input image source. It can be a file path, numpy array, or PIL Image.
            crop_padding (int, optional): Padding for cropping the detected tables. Defaults to 10.
            output_formats (List[OutputFormat], optional): Output formats for the reconstructed tables. Defaults to output all formats.
            border_format (BorderFormat, optional): Border format for the reconstructed tables. Defaults to BorderFormat.FULL_LINE.
            vertical_align (VerticalAlign, optional): Vertical align for the reconstructed tables. Defaults to VerticalAlign.MIDDLE.
            horizontal_align (HorizontalAlign, optional): Horizontal align for the reconstructed tables. Defaults to HorizontalAlign.LEFT.

        Returns:
            List[Dict]: List of detected tables with the reconstructed table string.
        """
        ## Step 1: Detect tables
        detected_tables = self.table_detection.process(
            image_source=image_source,
            conf=self.default_hparams["td"]["conf"],
            iou=self.default_hparams["td"]["iou"],
            extend_ratio=self.default_hparams["td"]["extend_ratio"],
            padding=self.default_hparams["td"]["padding"],
        )

        fail_tables = []

        ## Step 2: Extract table structure and values
        for table in detected_tables:

            ### Step 2.1: Table structure recognition
            table_image = table["table_image"]
            table_structure = self.table_structure_recognition.process(
                image_source=table_image,
                conf=self.default_hparams["tsr"]["conf"],
                iou=self.default_hparams["tsr"]["iou"],
            )

            if (
                table_structure is None
                or table_structure["num_rows"] == 0
                or table_structure["num_columns"] == 0
                or len(table_structure["cells"]) == 0
            ):
                fail_tables.append(table["table_id"])
                continue

            ### Step 2.2: Text extraction
            table_values = self.text_extraction.extract_text(table_image)

            ### Step 2.3: Table reconstruction
            table_outputs = self.table_reconstruction.process(
                table_structure=table_structure,
                table_values=table_values,
                output_formats=output_formats,
                border_format=border_format,
                vertical_align=vertical_align,
                horizontal_align=horizontal_align,
            )

            table["table_structure"] = table_structure
            table["table_values"] = table_values
            table["outputs"] = table_outputs["outputs"]

        ## Step 3: Remove fail tables
        detected_tables = [
            table for table in detected_tables if table["table_id"] not in fail_tables
        ]
        return detected_tables



from pathlib import Path
from typing import Union

import gdown
import torch
from ultralytics import YOLO

from ..const import MODEL_URLS, Task


def verify_device(device: Union[str, int]) -> torch.device:
    """Verify if the device is available.

    Args:
        device (Union[str, int]): "cpu" or cuda device id (int) or cuda device name (str). If not available, use "cpu".

    Returns:
        torch.device: device
    """
    if device != "cpu" and torch.cuda.is_available():
        all_device_ids = [i for i in range(torch.cuda.device_count())]

        if isinstance(device, str):
            device = int(device)

        if device not in all_device_ids:
            print(
                f"Device {device} is not available. Available devices are {all_device_ids}. Using CPU instead."
            )
            return torch.device("cpu")
        else:
            return torch.device(f"cuda:{device}")
    return torch.device("cpu")


def verify_weight(weight_path: str, task: Task, version: str) -> str:
    """Verify the weight path. If not exists, download the default weight. If failed to load, use the default weight instead.

    Args:
        weight_path (str): Path to the weight file.
        task (Task): Task to verify the weight.
        version (str): Version of the weight.

    Returns:
        str: Verified weight path.
    """
    ## Define default weight path
    default_weight_dir = Path("~/.yolo4tab/").expanduser()
    if not default_weight_dir.exists():
        default_weight_dir.mkdir(parents=True, exist_ok=True)
    default_weight_path = default_weight_dir.joinpath(MODEL_URLS[task][version]['file'])

    ## Download the default weight if not exists
    if not default_weight_path.exists():
        print(f"Downloading the weight for task '{task}'...")
        weight_url = MODEL_URLS[task][version]['url']
        gdown.download(weight_url, str(default_weight_path), quiet=False)
        print(
            f"Downloaded the weight for task '{task}' version '{version}' at '{default_weight_path}'."
        )

    ## Verify the weight path
    if weight_path is None:
        weight_path = default_weight_path
    else:
        try:
            _ = YOLO(weight_path)
        except Exception as e:
            print(
                f"Failed to load the weight at '{weight_path}'. Error: {e}. Using default weight instead."
            )
            weight_path = default_weight_path
    return weight_path



from fitz import Rect


def iob(bbox1, bbox2):
    """
    Compute the intersection area over box area, for bbox1.
    """
    intersection = Rect(bbox1).intersect(bbox2)

    bbox1_area = Rect(bbox1).get_area()
    if bbox1_area > 0:
        return intersection.get_area() / bbox1_area

    return 0


def sort_objects_by_score(objects, reverse=True):
    """
    Put any set of objects in order from high score to low score.
    """
    if reverse:
        sign = -1
    else:
        sign = 1
    return sorted(objects, key=lambda k: sign * k['score'])


def sort_objects_left_to_right(objs):
    """
    Put the objects in order from left to right.
    """
    return sorted(objs, key=lambda k: k['bbox'][0] + k['bbox'][2])


def sort_objects_top_to_bottom(objs):
    """
    Put the objects in order from top to bottom.
    """
    return sorted(objs, key=lambda k: k['bbox'][1] + k['bbox'][3])


def nms(
    objects, match_criteria="object2_overlap", match_threshold=0.05, keep_higher=True
):
    """
    A customizable version of non-maxima suppression (NMS).

    Default behavior: If a lower-confidence object overlaps more than 5% of its area
    with a higher-confidence object, remove the lower-confidence object.

    objects: set of dicts; each object dict must have a 'bbox' and a 'score' field
    match_criteria: how to measure how much two objects "overlap"
    match_threshold: the cutoff for determining that overlap requires suppression of one object
    keep_higher: if True, keep the object with the higher metric; otherwise, keep the lower
    """
    if len(objects) == 0:
        return []

    objects = sort_objects_by_score(objects, reverse=keep_higher)

    num_objects = len(objects)
    suppression = [False for obj in objects]

    for object2_num in range(1, num_objects):
        object2_rect = Rect(objects[object2_num]['bbox'])
        object2_area = object2_rect.get_area()
        for object1_num in range(object2_num):
            if not suppression[object1_num]:
                object1_rect = Rect(objects[object1_num]['bbox'])
                object1_area = object1_rect.get_area()
                intersect_area = object1_rect.intersect(object2_rect).get_area()
                try:
                    if match_criteria == "object1_overlap":
                        metric = intersect_area / object1_area
                    elif match_criteria == "object2_overlap":
                        metric = intersect_area / object2_area
                    elif match_criteria == "iou":
                        metric = intersect_area / (
                            object1_area + object2_area - intersect_area
                        )
                    if metric >= match_threshold:
                        suppression[object2_num] = True
                        break
                except Exception:
                    # Intended to recover from divide-by-zero
                    pass

    return [obj for idx, obj in enumerate(objects) if not suppression[idx]]


def apply_threshold(objects, threshold):
    """
    Filter out objects below a certain score.
    """
    return [obj for obj in objects if obj['score'] >= threshold]


def align_supercells(supercells, rows, columns):
    """
    For each supercell, align it to the rows it intersects 50% of the height of,
    and the columns it intersects 50% of the width of.
    Eliminate supercells for which there are no rows and columns it intersects 50% with.
    """
    aligned_supercells = []

    for supercell in supercells:
        row_bbox_rect = None
        col_bbox_rect = None
        intersecting_data_rows = set()
        for row_num, row in enumerate(rows):
            row_height = row['bbox'][3] - row['bbox'][1]
            supercell_height = supercell['bbox'][3] - supercell['bbox'][1]
            min_row_overlap = max(row['bbox'][1], supercell['bbox'][1])
            max_row_overlap = min(row['bbox'][3], supercell['bbox'][3])
            overlap_height = max_row_overlap - min_row_overlap
            if 'span' in supercell:
                overlap_fraction = max(
                    overlap_height / row_height, overlap_height / supercell_height
                )
            else:
                overlap_fraction = overlap_height / row_height
            if overlap_fraction >= 0.5:
                intersecting_data_rows.add(row_num)

        intersecting_rows = intersecting_data_rows
        # Determine vertical span of aligned supercell
        for row_num in intersecting_rows:
            if row_bbox_rect is None:
                row_bbox_rect = Rect(rows[row_num]['bbox'])
            else:
                row_bbox_rect = row_bbox_rect.include_rect(rows[row_num]['bbox'])
        if row_bbox_rect is None:
            continue

        intersecting_cols = []
        for col_num, col in enumerate(columns):
            col_width = col['bbox'][2] - col['bbox'][0]
            supercell_width = supercell['bbox'][2] - supercell['bbox'][0]
            min_col_overlap = max(col['bbox'][0], supercell['bbox'][0])
            max_col_overlap = min(col['bbox'][2], supercell['bbox'][2])
            overlap_width = max_col_overlap - min_col_overlap
            if 'span' in supercell:
                overlap_fraction = max(
                    overlap_width / col_width, overlap_width / supercell_width
                )
            else:
                overlap_fraction = overlap_width / col_width
            if overlap_fraction >= 0.5:
                intersecting_cols.append(col_num)
                if col_bbox_rect is None:
                    col_bbox_rect = Rect(col['bbox'])
                else:
                    col_bbox_rect = col_bbox_rect.include_rect(col['bbox'])
        if col_bbox_rect is None:
            continue

        supercell_bbox = list(row_bbox_rect.intersect(col_bbox_rect))
        supercell['bbox'] = supercell_bbox

        # Only a true supercell if it joins across multiple rows or columns
        if (
            len(intersecting_rows) > 0
            and len(intersecting_cols) > 0
            and (len(intersecting_rows) > 1 or len(intersecting_cols) > 1)
        ):
            supercell['row_numbers'] = list(intersecting_rows)
            supercell['column_numbers'] = intersecting_cols
            aligned_supercells.append(supercell)

    return aligned_supercells


def nms_supercells(supercells):
    """
    A NMS scheme for supercells that first attempts to shrink supercells to
    resolve overlap.
    If two supercells overlap the same (sub)cell, shrink the lower confidence
    supercell to resolve the overlap. If shrunk supercell is empty, remove it.
    """

    supercells = sort_objects_by_score(supercells)
    num_supercells = len(supercells)
    suppression = [False for supercell in supercells]

    for supercell2_num in range(1, num_supercells):
        supercell2 = supercells[supercell2_num]
        for supercell1_num in range(supercell2_num):
            supercell1 = supercells[supercell1_num]
            remove_supercell_overlap(supercell1, supercell2)
        if (
            (
                len(supercell2['row_numbers']) < 2
                and len(supercell2['column_numbers']) < 2
            )
            or len(supercell2['row_numbers']) == 0
            or len(supercell2['column_numbers']) == 0
        ):
            suppression[supercell2_num] = True

    return [obj for idx, obj in enumerate(supercells) if not suppression[idx]]


def remove_supercell_overlap(supercell1, supercell2):
    """
    This function resolves overlap between supercells (supercells must be
    disjoint) by iteratively shrinking supercells by the fewest grid cells
    necessary to resolve the overlap.
    Example:
    If two supercells overlap at grid cell (R, C), and supercell #1 is less
    confident than supercell #2, we eliminate either row R from supercell #1
    or column C from supercell #1 by comparing the number of columns in row R
    versus the number of rows in column C. If the number of columns in row R
    is less than the number of rows in column C, we eliminate row R from
    supercell #1. This resolves the overlap by removing fewer grid cells from
    supercell #1 than if we eliminated column C from it.
    """
    common_rows = set(supercell1['row_numbers']).intersection(
        set(supercell2['row_numbers'])
    )
    common_columns = set(supercell1['column_numbers']).intersection(
        set(supercell2['column_numbers'])
    )

    # While the supercells have overlapping grid cells, continue shrinking the less-confident
    # supercell one row or one column at a time
    while len(common_rows) > 0 and len(common_columns) > 0:
        # Try to shrink the supercell as little as possible to remove the overlap;
        # if the supercell has fewer rows than columns, remove an overlapping column,
        # because this removes fewer grid cells from the supercell;
        # otherwise remove an overlapping row
        if len(supercell2['row_numbers']) < len(supercell2['column_numbers']):
            min_column = min(supercell2['column_numbers'])
            max_column = max(supercell2['column_numbers'])
            if max_column in common_columns:
                common_columns.remove(max_column)
                supercell2['column_numbers'].remove(max_column)
            elif min_column in common_columns:
                common_columns.remove(min_column)
                supercell2['column_numbers'].remove(min_column)
            else:
                supercell2['column_numbers'] = []
                common_columns = set()
        else:
            min_row = min(supercell2['row_numbers'])
            max_row = max(supercell2['row_numbers'])
            if max_row in common_rows:
                common_rows.remove(max_row)
                supercell2['row_numbers'].remove(max_row)
            elif min_row in common_rows:
                common_rows.remove(min_row)
                supercell2['row_numbers'].remove(min_row)
            else:
                supercell2['row_numbers'] = []
                common_rows = set()


def refine_rows(rows, threshold=0.5):
    """
    Apply operations to the detected rows, such as
    thresholding, NMS, and alignment.
    """

    rows = nms(
        rows,
        match_criteria="object2_overlap",
        match_threshold=threshold,
        keep_higher=True,
    )
    if len(rows) > 1:
        rows = sort_objects_top_to_bottom(rows)

    return rows


def refine_columns(columns, threshold=0.5):
    """
    Apply operations to the detected columns, such as
    thresholding, NMS, and alignment.
    """
    columns = nms(
        columns,
        match_criteria="object2_overlap",
        match_threshold=threshold,
        keep_higher=True,
    )
    if len(columns) > 1:
        columns = sort_objects_left_to_right(columns)

    return columns


def align_columns(columns, bbox):
    """
    For every column, align the top and bottom boundaries to the final
    table bounding box.
    """
    try:
        for column in columns:
            column['bbox'][1] = bbox[1]
            column['bbox'][3] = bbox[3]
    except Exception as err:
        print("Could not align columns: {}".format(err))
        pass

    return columns


def align_rows(rows, bbox):
    """
    For every row, align the left and right boundaries to the final
    table bounding box.
    """
    try:
        for row in rows:
            row['bbox'][0] = bbox[0]
            row['bbox'][2] = bbox[2]
    except Exception as err:
        print("Could not align rows: {}".format(err))
        pass

    return rows


def refine_table_structure(table_structure, class_thresholds):
    """
    Apply operations to the detected table structure objects such as
    thresholding, NMS, and alignment.
    """
    rows = table_structure["rows"]
    columns = table_structure['columns']
    spanning_cells = table_structure['spanning_cells']

    # Process spanning cells
    spanning_cells = apply_threshold(
        spanning_cells, class_thresholds["table spanning cell"]
    )

    # Align before NMS for spanning cells because alignment brings them into agreement
    # with rows and columns first; if spanning cells still overlap after this operation,
    # the threshold for NMS can basically be lowered to just above 0
    spanning_cells = align_supercells(spanning_cells, rows, columns)
    spanning_cells = nms_supercells(spanning_cells)

    table_structure['columns'] = columns
    table_structure['rows'] = rows
    table_structure['spanning_cells'] = spanning_cells

    return table_structure



from typing import Dict, List, Union

import numpy as np
from fitz import Rect
from PIL import Image
from ultralytics import YOLO

from ..const import TSR_VERSION, Task
from ..utils.common import (
    align_columns,
    align_rows,
    iob,
    refine_columns,
    refine_rows,
    refine_table_structure,
)
from ..utils.general import verify_device, verify_weight


ID_TO_LABEL = {
    0: "table",
    1: "table row",
    2: "table column",
    3: "table spanning cell",
}

CLASS_THRESHOLD = {
    "table": 0.5,
    # "table column": 0.5,
    "table column": 0.25,
    "table row": 0.5,
    "table spanning cell": 0.5,
}


class TableStructureRecognition:
    def __init__(
        self,
        weight_path: str = None,
        device: str = "cpu",
    ):
        """Initialize TableStructureRecognition module

        Args:
            weight_path (str): Path to the weight file. Defaults to None (use default weight).
            device (str, optional): Device to run the model. Defaults to "cpu".
        """
        device = verify_device(device)
        self.weight_path = verify_weight(
            weight_path, Task.TABLE_STRUCTURE_RECOGNITION, TSR_VERSION
        )
        self.model = YOLO(self.weight_path).to(device)

    def process(
        self,
        image_source: Union[str, np.ndarray, Image.Image],
        conf: float = 0.25,
        iou: float = 0.7,
    ) -> Dict:
        """Perform table structure recognition on the input image source

        Args:
            image_source (Union[str, np.ndarray, Image.Image]): Input image source. Can be path to image file, numpy array or PIL Image.
            conf (float, optional): Confidence threshold. Defaults to 0.25.
            iou (float, optional): IOU threshold. Defaults to 0.7.

        Returns:
            Dict: A dictionary containing table structure with keys: table, rows, columns, spanning_cells
        """

        infer_results = self._infer(image_source=image_source, conf=conf, iou=iou)

        table_objects = []
        for tb_object in infer_results:

            object_label = ID_TO_LABEL[int(tb_object.boxes.cls)]
            object_bbox = tb_object.boxes.xyxy[0].int().tolist()
            object_score = float(tb_object.boxes.conf)

            object_result = {
                "label": object_label,
                "bbox": object_bbox,
                "score": object_score,
            }

            table_objects.append(object_result)

        ## Post process to get table structure
        if len(table_objects) == 0:
            return None
        table_structure = self._post_process(table_objects)
        if (
            table_structure is None
            or len(table_structure["rows"]) == 0
            or len(table_structure["columns"]) == 0
        ):
            return None
        table_structure = self._get_cells(table_structure)
        return table_structure

    def _infer(
        self,
        image_source: Union[str, np.ndarray, Image.Image],
        conf: float = 0.25,
        iou: float = 0.7,
    ) -> List:
        """Perform inference on the input image source

        Args:
            image_source (Union[str, np.ndarray, Image.Image]): Input image source. Can be path to image file, numpy array or PIL Image.
            conf (float, optional): Confidence threshold. Defaults to 0.25.
            iou (float, optional): IOU threshold. Defaults to 0.7.

        Returns:
            List: List of result object defined by Ultralytics
        """
        results = self.model.predict(
            source=image_source, conf=conf, iou=iou, verbose=False
        )
        return results[0]

    def _post_process(
        self, table_objetcs: List[Dict], iob_threshold: float = 0.5
    ) -> Dict:
        """Post process to get table structure

        Args:
            table_objetcs (List[Dict]): List of table objects detected by the model with keys: label, bbox, score.
            iob_threshold (float, optional): IOB threshold. Defaults to 0.5.

        Returns:
            Dict: A dictionary containing table structure with keys: table, rows, columns, spanning_cells.
        """
        ## Step 1 -> Get table bbox
        table = [obj for obj in table_objetcs if obj["label"] == "table"]

        ### If there is no table detected, return None
        if len(table) == 0:
            return None

        ### Else, get the table with highest score
        table = sorted(table, key=lambda x: x['score'], reverse=True)
        table = table[0]

        ## Step 2 -> Get objects in table which iob > iob_threshold
        ## and classifiy them to rows, columns, spanning cells
        table_objects = [
            obj
            for obj in table_objetcs
            if iob(obj['bbox'], table['bbox']) >= iob_threshold
        ]

        ## Step 3: Get row, columns and spaning cells
        rows, columns, spaning_cells = [], [], []
        for obj in table_objects:
            if obj["label"] == "table row":
                rows.append(obj)
            elif obj["label"] == "table column":
                columns.append(obj)
            elif obj["label"] == "table spanning cell":
                spaning_cells.append(obj)

        ### If no rows/columns left, add a row/column based on table bbox
        if len(rows) == 0:
            rows.append(
                {
                    "label": "table row",
                    "bbox": table['bbox'],
                    "score": 1.0,
                }
            )

        if len(columns) == 0:
            columns.append(
                {
                    "label": "table column",
                    "bbox": table['bbox'],
                    "score": 1.0,
                }
            )

        ## Step 4 -> Refine table structures (rows, columns, spanning cells) by NMS threshold
        rows = refine_rows(rows, CLASS_THRESHOLD["table row"])
        columns = refine_columns(columns, CLASS_THRESHOLD["table column"])

        ## Step 5 -> Align rows and columns to fit table bounding box
        row_rect = Rect()
        for obj in rows:
            row_rect.include_rect(obj['bbox'])

        column_rect = Rect()
        for obj in columns:
            column_rect.include_rect(obj['bbox'])

        table['row_column_bbox'] = [
            column_rect[0],
            row_rect[1],
            column_rect[2],
            row_rect[3],
        ]
        table['bbox'] = table['row_column_bbox']

        columns = align_columns(columns, table['row_column_bbox'])
        rows = align_rows(rows, table['row_column_bbox'])

        ## Step 6 -> Get structure
        table_structure = {}
        table_structure['table'] = table
        table_structure['rows'] = rows
        table_structure['columns'] = columns
        table_structure['spanning_cells'] = spaning_cells

        ## Step 7 -> Refine table structure
        if len(rows) > 0 and len(columns) > 1:
            table_structure = refine_table_structure(table_structure, CLASS_THRESHOLD)

        ## Step 8 -> Save result
        return table_structure

    def _get_cells(self, table_structure: Dict) -> List[Dict]:
        """Get list of cells from detected table structure

        Args:
            table_structure (Dict): Table components detected and refined. Dictionary with keys: table, rows, columns, spanning_cells.

        Returns:
            Dict: List of cells in table. Cells are represented by dictionary with keys:
                - bbox: Bounding box of cell
                - column_nums: List of column numbers that cell spans
                - row_nums: List of row numbers that cell spans
        """

        columns = table_structure["columns"]
        rows = table_structure["rows"]
        num_rows = len(rows)
        num_columns = len(columns)
        spanning_cells = table_structure["spanning_cells"]
        cells = []
        subcells = []

        ## Identify complete cells and subcells
        for column_num, column in enumerate(columns):
            for row_num, row in enumerate(rows):
                column_rect = Rect(list(column['bbox']))
                row_rect = Rect(list(row['bbox']))
                cell_rect = row_rect.intersect(column_rect)
                cell = {
                    'bbox': list(cell_rect),
                    'column_nums': [column_num],
                    'row_nums': [row_num],
                }

                cell['subcell'] = False
                for spanning_cell in spanning_cells:
                    spanning_cell_rect = Rect(list(spanning_cell['bbox']))
                    if (
                        spanning_cell_rect.intersect(cell_rect).get_area()
                        / cell_rect.get_area()
                    ) > 0.5:
                        cell['subcell'] = True
                        break

                if cell['subcell']:
                    subcells.append(cell)
                else:
                    cells.append(cell)

        for spanning_cell in spanning_cells:
            spanning_cell_rect = Rect(list(spanning_cell['bbox']))
            cell_columns = set()
            cell_rows = set()
            cell_rect = None
            for subcell in subcells:
                subcell_rect = Rect(list(subcell['bbox']))
                subcell_rect_area = subcell_rect.get_area()
                if (
                    subcell_rect.intersect(spanning_cell_rect).get_area()
                    / subcell_rect_area
                ) > 0.5:
                    if cell_rect is None:
                        cell_rect = Rect(list(subcell['bbox']))
                    else:
                        cell_rect.include_rect(Rect(list(subcell['bbox'])))
                    cell_rows = cell_rows.union(set(subcell['row_nums']))
                    cell_columns = cell_columns.union(set(subcell['column_nums']))

            if len(cell_rows) > 0 and len(cell_columns) > 0:
                cell = {
                    'bbox': list(cell_rect),
                    'column_nums': list(cell_columns),
                    'row_nums': list(cell_rows),
                }
                cells.append(cell)

        dilated_columns = columns
        dilated_rows = rows
        for cell in cells:
            column_rect = Rect()
            for column_num in cell['column_nums']:
                column_rect.include_rect(list(dilated_columns[column_num]['bbox']))
            row_rect = Rect()
            for row_num in cell['row_nums']:
                row_rect.include_rect(list(dilated_rows[row_num]['bbox']))
            cell_rect = column_rect.intersect(row_rect)
            cell['bbox'] = list(cell_rect)

        ## Sort cells from left to right, top to bottom
        sorted_cells = sorted(
            cells, key=lambda cell: (cell['row_nums'][0], cell['column_nums'][0])
        )

        ## Drop 'subcell' key
        for cell in sorted_cells:
            cell.pop('subcell', None)

        table_structure['cells'] = sorted_cells
        table_structure['num_rows'] = num_rows
        table_structure['num_columns'] = num_columns
        return table_structure



from .const import *
from .table_detection import TableDetection
from .table_extraction import TableExtraction
from .table_reconstruction import TableReconstruction
from .table_structure_recognition import TableStructureRecognition
from .text_extraction import TextExtraction
from .utils import *


__version__ = "0.2.1"

__all__ = [
    TableExtraction,
    TableDetection,
    TableStructureRecognition,
    TextExtraction,
    TableReconstruction,
]



from typing import Dict, List, Union
from uuid import uuid4

import cv2
import numpy as np
from PIL import Image
from ultralytics import YOLO

from ..const import TD_VERSION, Task
from ..utils.general import verify_device, verify_weight
from ..utils.warp import (
    crop_rotate_bbox,
    crop_segmentation,
    extend_polygon,
    find_approximate_polygon,
    find_bbox,
)


class TableDetection:
    def __init__(
        self,
        weight_path: str = None,
        device: str = "cpu",
    ):
        """Initialize TableDetection module

        Args:
            weight_path (str): Path to the weight file. Defaults to None (use default weight).
            device (str, optional): Device to run the model. Defaults to "cpu".
        """
        device = verify_device(device)
        self.weight_path = verify_weight(weight_path, Task.TABLE_DETECTION, TD_VERSION)
        self.model = YOLO(self.weight_path).to(device)

    def process(
        self,
        image_source: Union[str, np.ndarray, Image.Image],
        conf: float = 0.25,
        iou: float = 0.7,
        extend_ratio: float = 0.1,
        padding: int = 10,
    ) -> List[Dict]:
        """Perform table detection on the input image source

        Args:
            image_source (Union[str, np.ndarray, Image.Image]): Input image source. Can be path to image file, numpy array or PIL Image.
            conf (float, optional): Confidence threshold. Defaults to 0.25.
            iou (float, optional): IOU threshold. Defaults to 0.7.
            extend_ratio (float, optional): Ratio for extending the detected table based on smallest dimension of the table bbox. Defaults to 0.1.
            padding (int, optional): Size of white padding for cropping the table. Defaults to 10.

        Returns:
            List[Dict]: List of detected tables. Each table is a dictionary with keys: table_id, bbox, mask, score, ultralytics, table_image, approx_poly
        """
        infer_results = self._infer(image_source, conf=conf, iou=iou)

        final_results = []
        for table in infer_results:
            table_bbox = table.boxes.xyxy[0].int().tolist()
            table_segment = table.masks.xy[0].astype(int)
            score = table.boxes.conf.item()
            table_result = {
                "table_id": str(uuid4()),
                "bbox": table_bbox,
                "mask": table_segment,
                "score": score,
                "ultralytics": table,
                "table_image": None,
            }

            if table_segment is not None and table_bbox is not None:
                orig_img = table.orig_img
                height, width = table.orig_shape

                detect_width = table_bbox[2] - table_bbox[0]
                detect_height = table_bbox[3] - table_bbox[1]

                approx_poly = find_approximate_polygon(table_segment)
                if len(approx_poly) == 4:
                    approx_poly = extend_polygon(
                        approx_poly,
                        width,
                        height,
                        padding=extend_ratio * min(detect_width, detect_height),
                    )
                    crop_img, _ = crop_segmentation(orig_img, approx_poly)
                else:
                    approx_bbox = find_bbox(table_segment)
                    approx_bbox = extend_polygon(
                        approx_bbox,
                        width,
                        height,
                        padding=extend_ratio * min(detect_width, detect_height),
                    )
                    crop_img, _ = crop_rotate_bbox(orig_img, approx_bbox)

                crop_img = cv2.copyMakeBorder(
                    crop_img,
                    padding,
                    padding,
                    padding,
                    padding,
                    cv2.BORDER_CONSTANT,
                    value=(255, 255, 255),
                )
                table_result["approx_poly"] = approx_poly
                table_result["table_image"] = crop_img.astype(np.uint8)

            final_results.append(table_result)

        ## Sort by y0 of table bbox, then x0 of table bbox
        final_results = sorted(
            final_results, key=lambda x: (x["bbox"][1], x["bbox"][0])
        )

        return final_results

    def _infer(
        self,
        image_source: Union[str, np.ndarray, Image.Image],
        conf: float = 0.25,
        iou: float = 0.7,
    ) -> List:
        """Perform inference on the input image source

        Args:
            image_source (Union[str, np.ndarray, Image.Image]): Input image source. Can be path to image file, numpy array or PIL Image.
            conf (float, optional): Confidence threshold. Defaults to 0.25.
            iou (float, optional): IOU threshold. Defaults to 0.7.

        Returns:
            List: List of result object defined by Ultralytics
        """
        results = self.model.predict(
            source=image_source, conf=conf, iou=iou, retina_masks=True, verbose=False
        )
        return results[0]



import os
from typing import List

import cv2
import numpy as np
import torch
from tqdm import tqdm
from ultralytics import YOLO


def crop_rotate_bbox(img: np.ndarray, corners: np.ndarray) -> np.ndarray:
    corners = corners.astype(np.float32)
    assert len(corners) == 4, "Shape of points must be 4*2"
    img_crop_width = int(
        max(
            np.linalg.norm(corners[0] - corners[1]),
            np.linalg.norm(corners[2] - corners[3]),
        )
    )
    img_crop_height = int(
        max(
            np.linalg.norm(corners[0] - corners[3]),
            np.linalg.norm(corners[1] - corners[2]),
        )
    )
    pts_std = np.float32(
        [
            [0, 0],
            [img_crop_width, 0],
            [img_crop_width, img_crop_height],
            [0, img_crop_height],
        ]
    )
    M = cv2.getPerspectiveTransform(corners, pts_std)
    img_crop = cv2.warpPerspective(
        img,
        M,
        (img_crop_width, img_crop_height),
        borderMode=cv2.BORDER_REPLICATE,
        flags=cv2.INTER_CUBIC,
    )
    return img_crop, M


def crop_segmentation(img: np.ndarray, corners: np.ndarray):
    # Sorting follow ABCD
    x_sorted = sorted(corners, key=lambda point: point[0])
    left_most = x_sorted[:2]
    right_most = x_sorted[2:]

    left_most = sorted(left_most, key=lambda point: point[1])
    right_most = sorted(right_most, key=lambda point: point[1])

    quad_points = np.array(
        [left_most[0], right_most[0], right_most[1], left_most[1]], dtype=np.float32
    )

    # Find rectangle transform
    img_crop_width = np.linalg.norm(quad_points[0] - quad_points[1]).astype(np.int32)
    img_crop_height = np.linalg.norm(quad_points[0] - quad_points[3]).astype(np.int32)
    rect_points = np.array(
        [
            [0, 0],
            [img_crop_width, 0],
            [img_crop_width, img_crop_height],
            [0, img_crop_height],
        ],
        np.float32,
    )

    # Get matrix M and transform image
    M = cv2.getPerspectiveTransform(quad_points, rect_points)
    img_crop = cv2.warpPerspective(
        img.copy().astype(np.float32), M, (img_crop_width, img_crop_height)
    )

    return img_crop, M


def find_approximate_polygon(segment, epsilon_ratio=0.016):
    epsilon = epsilon_ratio * cv2.arcLength(segment, True)
    approx = cv2.approxPolyDP(segment, epsilon, True)
    approx = approx.reshape(-1, 2)

    return approx


def find_bbox(segment):
    bounding_box = cv2.minAreaRect(segment)
    points = sorted(cv2.boxPoints(bounding_box).tolist(), key=lambda x: x[0])

    if points[1][1] > points[0][1]:
        index_1 = 0
        index_4 = 1
    else:
        index_1 = 1
        index_4 = 0
    if points[3][1] > points[2][1]:
        index_2 = 2
        index_3 = 3
    else:
        index_2 = 3
        index_3 = 2

    # return a numpy array of shape (4, 2) with the order of points is top-left, top-right, bottom-right, bottom-left
    return np.array(
        [points[index_1], points[index_2], points[index_3], points[index_4]]
    )


def extend_polygon(polygon, img_width, img_height, padding=20):
    center = np.mean(polygon, axis=0)
    vectors = polygon - center

    lengths = np.linalg.norm(vectors, axis=1)
    scaled_vectors = vectors + vectors * (padding / lengths)[:, np.newaxis]
    expanded_polygon = center + scaled_vectors

    expanded_polygon = np.reshape(expanded_polygon, (-1, 2)).astype(np.int32)

    for p in expanded_polygon:
        p[0] = min(max(p[0], 0), img_width - 1)
        p[1] = min(max(p[1], 0), img_height - 1)

    return expanded_polygon


if __name__ == "__main__":
    # Load a model
    weight = "/home/manhckv/manhckv/ultralytics/runs/train/exp1/weights/best.pt"
    model = YOLO(weight)

    # test_img_dir = "/home/manhckv/manhckv/ultralytics/test"
    test_img_dir = "/home/manhckv/manhckv/ultralytics/td_test"
    # test_img_dir = "/home/manhckv/manhckv/ultralytics/fqa_test"

    output_dir = "_output"
    os.makedirs(output_dir, exist_ok=True)

    for img_file in tqdm(os.listdir(test_img_dir)):
        img_path = os.path.join(test_img_dir, img_file)
        img = cv2.imread(img_path)

        pred = model.predict(img_path, device="3", conf=0.6, iou=0.7)

        result = pred[0]

        if result.masks is None:
            print(f"No object detected in {img_file}")
            continue

        mask = result.masks.xy
        # img = visualize_segments(img, mask)

        corners = []

        crop_imgs = []

        for segment in mask:
            approx_poly = find_approximate_polygon(
                segment
            )  # np.array with shape (4, 2)
            if len(approx_poly) == 4:
                # when can find 4 corners, we can use it directly
                approx_poly = extend_polygon(approx_poly, img.shape[1], img.shape[0])
                corners.append(approx_poly)
                crop_img, _ = crop_segmentation(img, approx_poly)
                crop_imgs.append(crop_img)

            else:
                # else, we need to find the bounding box of the segment
                approx_bbox = find_bbox(segment)
                approx_bbox = extend_polygon(approx_bbox, img.shape[1], img.shape[0])
                corners.append(approx_bbox)
                crop_img, _ = crop_segmentation(img, approx_bbox)
                crop_imgs.append(crop_img)

        # img = plot_4_corner(img, corners)
        # cv2.imwrite(os.path.join(output_dir, img_file), img)
        for idx, crop_img in enumerate(crop_imgs):
            cv2.imwrite(os.path.join(output_dir, f"{img_file}_{idx}.jpg"), crop_img)



from enum import auto

from strenum import LowercaseStrEnum


TD_VERSION = "v0.0.2"
TSR_VERSION = "v0.0.1"

MODEL_URLS = {
    "table_detection": {
        "v0.0.1": {
            "url": "https://drive.google.com/uc?id=1_owQ6JvJkEVCS4nhkmaTE8sIBDj6aQwF",
            "file": "td_v0.0.1.pt",
        },
        "v0.0.2": {
            "url": "https://drive.google.com/uc?id=1FdAAw9CqV9popKyqbXcslqn7tYyJLmn8",
            "file": "td_v0.0.2.pt",
        },
    },
    "table_structure_recognition": {
        "v0.0.1": {
            "url": "https://drive.google.com/uc?id=14sXs0Z6zKDeu41uKStl44hxfM1ghhg0d",
            "file": "tsr_v0.0.1.pt",
        }
    },
}


class Task(LowercaseStrEnum):
    TABLE_DETECTION = auto()
    TABLE_STRUCTURE_RECOGNITION = auto()


class OutputFormat(LowercaseStrEnum):
    HTML = auto()
    LATEX = auto()
    CSV = auto()


class BorderFormat(LowercaseStrEnum):
    FULL_BORDER = auto()
    NO_BORDER = auto()
    HOR_BORDER = auto()
    VER_BORDER = auto()
    INNER_BORDER = auto()
    OUTER_BORDER = auto()


class VerticalAlign(LowercaseStrEnum):
    TOP = auto()
    MIDDLE = auto()
    BOTTOM = auto()


class HorizontalAlign(LowercaseStrEnum):
    LEFT = auto()
    CENTER = auto()
    RIGHT = auto()
